open Lambda
open Asttypes
open Typedtree
open Transltools

let debug = try ignore (Sys.getenv "DEBUG_TRANSLFLOW"); true with _ -> false

let transl_flow env flow =
  if debug then Typegeneric.print_flow flow;
  let loopback_table =
    let rec get_loops = function
      |	Flow_node node -> List.flatten (List.map get_loops node.flow_sub)
      |	Flow_loop id -> [id, ref None]
      |	Flow_type _ -> []
    in
    get_loops flow
  in
  
  let rec sub = function
    | Flow_node node ->
	begin try
	  let idref = List.assoc node.flow_id loopback_table in
	  (* require let rec *)
	  let v = Ident.create "loop" in
	  idref := Some v;
	  (* make_block 0 [Lletrec ([v, transl_node node], Lvar v)] *)
	  Lletrec ([v, transl_node node], Lvar v)
    	with
	| Not_found ->
	    (* make_block 0 [transl_node node] *)
	    transl_node node
	end
    | Flow_loop id ->
	begin match !(List.assoc id loopback_table) with
	| Some v -> (* make_block 0 [Lvar v] *) Lvar v
	| None -> assert false
	end
    | Flow_type typs ->
	(* List.map (fun typ ->
	   make_block 1 [Transltype.transl_rtype_of_type env typ]) typs *)
	make_block 0 (List.map (Transltype.transl_rtype_of_type env) typs)

  and transl_node node =
    make_block 0 ( make_int node.flow_pos (* flow_pos *) ::
                  List.map sub node.flow_sub) (* flow_sub *)
                   (* flow_id is not required *)
  in
  
  sub flow
;;
