(* dynamics *)

open Rdigest

exception Type_match_failure of (rpath * type_info) rtype * string * int * int

let type_of (v : dyn) = 
  let _, t = (Obj.magic v : 'a * (rpath * type_info) rtype) in
  t

let is_instance g t =
  let rec is_instance_aux sub g t =
    match g, t with
      _, RTvar i when i < 0 -> [] (* _ i.e. "any" type pattern *)
    | RTvar i, _ ->
  	begin 
  	  try
  	    if List.assoc i sub = t then sub else raise Exit
  	  with
  	    Not_found -> (i, t) :: sub
  	end
    | RTarrow (f1, t1), RTarrow (f2, t2) ->
  	let newsub = is_instance_aux sub f1 f2 in
  	is_instance_aux newsub t1 t2
    | RTtuple tl1, RTtuple tl2 when List.length tl1 = List.length tl2 -> 
  	List.fold_left2 (fun st t1 t2 -> is_instance_aux st t1 t2) sub tl1 tl2
    | RTconstr (i1, tl1), RTconstr (i2, tl2) ->
	if i1 <> i2 then raise Exit;
	if List.length tl1 <> List.length tl2 then raise Exit;
  	List.fold_left2 (fun st t1 t2 -> is_instance_aux st t1 t2) sub tl1 tl2
    | _ -> raise Exit
  in
  try
    is_instance_aux [] g t; true
  with
    Exit -> false

let dynamic_comp = fun ty v -> (v, ty)
let fail t m s e = raise (Type_match_failure ( t, m, s, e ))
let coerce_comp = fun (m,ls,le) ty1 (v, ty2) ->
  if is_instance ty2 ty1 then v
  else fail ty2 m ls le

(*** Generic primitives *)

external dyn : $a -> dyn = "%dynamic" "generic" "dynamic"
external crc : dyn -> $a = "%coerce" "generic" "coerce"

external export_value : out_channel -> $a -> unit = "%export_value" "generic" "dynamic"
external import_value : in_channel -> $a = "%import_value" "generic" "coerce"

external export_to_string : $a -> string = "%export_to_string" "generic" "dynamic"
external import_from_string : string -> $a = "%import_from_string" "generic" "coerce"
