open Misc
open Asttypes
open Types
open Btype
open Ctype

exception Generic_unsupported_error of string
exception Generic_system_error of string

(* anti-unification *)

val anti_unify :
  Env.t -> Types.type_expr -> Types.type_expr -> Types.type_expr
val anti_unify_types :
  Env.t -> Types.type_expr list -> Types.type_expr 

(* They may be replaced by Btype.free_vars *)
val normal_variables : type_expr -> type_expr list
        (* get the normal type variables occur in a type expression *)
val dynamic_variables : type_expr -> type_expr list
        (* get the dynamic variables occur in a type expression *)
val type_variables : type_expr -> type_expr list
        (* get the variables (normal/dynamic) occur in a type expression *)

val generalized_dynamic_variables : type_expr -> type_expr list
        (* get the generalized_dynamic variables occur in a type expression *)

val equal_scheme : Env.t -> type_expr -> type_expr -> bool
  
(* GENERIC: generalize_copy 
     Read the comment for the Tdvar in types.mli.
*)
val generalize_copy: type_expr -> type_expr
val force_generalize_copy: type_expr -> type_expr
val generalized_variables : unit -> (type_expr * type_expr) list
        (* Generalize the given type *)
        (* The generalized variables are created newly *)
        (* The original type variables to be generalized are still as are.
	   generalized_variables reports such variables. *) 
        (* in the format of (original variable, generalized variable) *)

type mode_for_links = Keep | Duplicate | Reset
type copy_cache
val create_copy_cache : mode_for_links -> copy_cache
val copy_type : copy_cache -> type_expr -> type_expr
val copy_types : mode_for_links -> type_expr list -> type_expr list

type size_dvar = {
    info: Types.tdvar_info ref;
    mutable num: int
  } 

type size = {
    mutable tdvars: size_dvar list;
    mutable const: int
  }

val size : type_expr -> size
val size_of_instance_predicate : 
    Env.t -> type_expr -> type_expr -> size

val leq : size -> size -> bool
val string_of_size : size -> string
